--[[
Server Version: 0.3.4
Author: 		Nathaniel Staples
Last edited by: Nathaniel Staples
Edited on: 		2016.07.17

Dependencies:
NONE

Purpose:
Adds additional messages to the combat script. The messages are sent to the weapon
and contain the following information:
	success (string) - possible values hit, parried, blocked, missed, dodged 
	attacker (object)
	defender (object)
	isCrit	(bool)
	damageAmmount
	
	weapon:SendMessage(WeaponHitMessage, success, attacker, defender, isCrit, damage)
	
Use:
File is written to be required in a dummy combat.lua script on a server using
Yorlik's script to enable folders for storing scripts that will be required.

--]]


--[[
atTarget = <OBJECT>  Target Object attack is being made against
specOp = <STRING> Special Ability being used to perform attack
wepHand = <STRING> Slot the weapon is being held in.
rangedAttack = <BOOL> is this a ranged attack or not (affects countering)

--]]

function ExecuteMissActions(atTarget,specOp,wepHand, rangedAttack)
	TDebugFunction("DebugMiss","FN_ ExecuteMissActions: " ..this:GetName() .. " -Missed- " .. atTarget:GetName())
	if(rangedAttack == nil) then rangedAttack = false end
	atTarget:NpcSpeech("[08FFFF]*Miss*[-]","combat")
	atTarget:SendMessage("AttackerMissedMessage", {["attacker"] = this, ["rangedAttack"] = rangedAttack})
	if(specOp == "MagicAttack") or (rangedAttack == true) then return end
	PlayWeaponSound(this,"Miss", this:GetEquippedObject(wepHand))
	local oDuration = GetStatusInfo("Vulnerable", "Duration", this) or 3 
	this:ScheduleTimerDelay(TimeSpan.FromSeconds(oDuration), "Vulnerable")


	--MOD CODE HERE
	local myWep = mCombatInfo[wepHand].object
	if (myWep) then
		myWep:SendMessage("WeaponHitMessage", "missed", this, atTarget, false, 0)
	end
	--MODE CODE HERE
end


function ExecuteParryActions(atTarget,specOp,wepHand, rangedAttack)
	TDebugFunction("DebugParry","FN_ ExecuteParryActions: " ..this:GetName() .. " was parried by " .. atTarget:GetName())
	atTarget:PlayAnimation("swing")
	atTarget:NpcSpeech("[0ab4f7]*Parry*[-]","combat")
	atTarget:SendMessage("ParriedAttackMessage", {["attacker"] = this, ["rangedAttack"] = rangedAttack})
	atTarget:SendMessage("AddFreezeEffectMessage", "Parry", 250)
	if(rangedAttack ~= true) then
		PlayWeaponSound(this,"Parry",this:GetEquippedObject(wepHand))
		local oDuration = GetStatusInfo("Open", "Duration", this)
		this:ScheduleTimerDelay(TimeSpan.FromSeconds(oDuration), "Open")
	end
	local swingResetPenalty = GetCombatActionInfo("AttackParried", "AttackerNextSwingPenalty",  this)
	--ResetSwingTimer(swingResetPenalty, wepHand)


	--MOD CODE HERE
	local myWep = mCombatInfo[wepHand].object
	if (myWep) then
		myWep:SendMessage("WeaponHitMessage", "parried", this, atTarget, false, 0)
	end
	--MODE CODE HERE
end


function ExecuteDodgeActions(atTarget,specOp,wepHand, rangedAttack)
	TDebugFunction("DebugDodge","FN_ ExecuteDodgeActions: " ..this:GetName() .. " was dodged by " .. atTarget:GetName())
	atTarget:SendMessage("DodgedAttackMessage", {["attacker"] = this, ["rangedAttack"] = rangedAttack})
	atTarget:PlayAnimation("dodge")
	atTarget:SendMessage("AddFreezeEffectMessage", "Dodge", 250)
	atTarget:NpcSpeech("[BA0AF7]*Dodge*[-]","combat")
	if(rangedAttack ~= true) then
		PlayWeaponSound(this,"Miss",this:GetEquippedObject(wepHand))
		local oDuration = GetStatusInfo("Open", "Duration", this) or 2
		this:ScheduleTimerDelay(TimeSpan.FromSeconds(oDuration), "Open")
	end
	
	local swingResetPenalty = GetCombatActionInfo("AttackDodged", "AttackerNextSwingPenalty",  this)
	--ResetSwingTimer(swingResetPenalty, wepHand)

	--MOD CODE HERE
	local myWep = mCombatInfo[wepHand].object
	if (myWep) then
		myWep:SendMessage("WeaponHitMessage", "dodged", this, atTarget, false, 0)
	end
	--MODE CODE HERE

end


function ExecuteBlockActions(atTarget,specOp,wepHand, rangedAttack)
	TDebugFunction("FN_ ExecuteBlockActions: " ..this:GetName() .. " was blocked by " .. atTarget:GetName())
	atTarget:SendMessage("BlockedAttackMessage", {["attacker"] = this, ["rangedAttack"] = rangedAttack})
	local oDuration = GetStatusInfo("Open", "Duration", this)
	this:ScheduleTimerDelay(TimeSpan.FromSeconds(oDuration), "Open")
	local swingResetPenalty = GetCombatActionInfo("AttackBlocked", "AttackerNextSwingPenalty",  this)
	--ResetSwingTimer(swingResetPenalty, wepHand)
	atTarget:PlayAnimation("block")
	atTarget:SendMessage("AddFreezeEffectMessage", "Block", 300)
	atTarget:NpcSpeech("[FCF403]*Blocked*[-]","combat")
	local blockDamageInfo = GetWeaponHitInfo(wepHand, mCombatInfo, nil)
	local baseDam = blockDamageInfo.damageAmt
	local blockDamageModifier = GetBlockDamageModifier(atTarget) or .5
	local damToApply = math.ceil(blockDamageModifier * baseDam)
	local myWep = mCombatInfo[wepHand].object
	if(myWep ~= nil) then
		myWep:SendMessage("WeaponBlocked", { ["victim"] = atTarget, ["damageDealt"] = damToApply, ["resisted"] = baseDam})
	end
	atTarget:SendMessage("DamageInflicted",this,damToApply,false, baseDam)
	local damagerOwner = this:GetObjVar("owner")
	CreditDamageDealt(this,atTarget,damToApply,baseDam,damagerOwner)

	--MOD CODE HERE
	if (myWep) then
		myWep:SendMessage("WeaponHitMessage", "blocked", this, atTarget, false, damToApply)
	end
	--MODE CODE HERE
end


function ExecuteRangedBlockActions(atTarget,damInfo)
	TDebugFunction("FN_ ExecuteBlockActions: " ..this:GetName() .. " was blocked by " .. atTarget:GetName())
	atTarget:SendMessage("BlockedAttackMessage", {["attacker"] = this, ["rangedAttack"] = true})
	atTarget:PlayAnimation("block")
	atTarget:SendMessage("AddFreezeEffectMessage", "Block", 300)
	atTarget:NpcSpeech("[FCF403]*Blocked*[-]","combat")
	PlayShieldSound(atTarget,"Block")
	local baseDam = damInfo.damageAmt
	local blockDamageModifier = GetBlockDamageModifier(atTarget) or .5
	local damToApply = math.ceil(blockDamageModifier * baseDam)
	atTarget:SendMessage("DamageInflicted",this,damToApply,false, baseDam)
	local damagerOwner = this:GetObjVar("owner")
	CreditDamageDealt(this,atTarget,damToApply,baseDam,damagerOwner)

	--MOD CODE HERE
	local myWep = mCombatInfo[wepHand].object
	if (myWep) then
		myWep:SendMessage("WeaponHitMessage", "blocked", this, atTarget, false, damToApply)
	end
	--MODE CODE HERE
	
end


function ExecuteHitActions(atTarget,specOp,wepHand, rangedAttack, instantAttack, facingZone)
	TDebugFunction("FN_ ExecuteHitActions: " ..this:GetName() .. " *Hit* " .. atTarget:GetName())
	--DebugMessage(tostring(this) .. " ****" .. tostring(wepHand))
	local selfFXBonus, selfFXMods = this:GetScriptStatMods("WeaponDamageMods")
	local meleeBonus = 0
	local meleeMods = 1
	--DebugMessage(mCombatInfo[wepHand])
	local wepDamSkill = mCombatInfo[wepHand].DamageSkill
	if(rangedAttack) then
		wepDamSkill = mCombatInfo[wepHand].ProjectileDamageSkill
	else
		meleeBonus, meleeMods = this:GetScriptStatMods("DamageMod")
	end
	local facingMod = GetFacingZoneModifier("MeleeDamageModifier", this, atTarget, facingZone)
	
	local specDamMod = 1
	if(specOp ~= nil) then
		if(instantAttack) then
			specDamMod = GetInstantDamageMod(specOp)
		else
			specDamMod = GetTriggeredDamageModifier(specOp)
		end
	end
	if((atTarget:HasObjVar("attackable") and atTarget:HasObjVar("CombatMaxSkillGain")) or (atTarget:IsMobile() and IsInCombat(atTarget) and not (IsDead(atTarget) or (atTarget:HasObjVar("Invulnerable") and atTarget:GetObjVar"Invulnerable")))) then
		local mySkill = this:GetSkillLevel(wepDamSkill)
		local maxSkillGain = atTarget:GetObjVar("CombatMaxSkillGain") or (mySkill + 1)
		if(maxSkillGain > mySkill) then
			AddCombatSkillGainCheck(wepDamSkill,this, atTarget)
		end
	end		

	local damInfo = GetWeaponHitInfo(wepHand, mCombatInfo, specDamMod * selfFXMods * meleeMods * facingMod)
	--DebugMessage(this:GetName() .. " Damage Modifer:" .. (specDamMod * selfFXMods) .. " Bonus: " .. (selfFXBonus + meleeBonus))
	local isCrit = CheckCritSuccess(atTarget, specOp,wepHand, rangedAttack)
	if(isCrit) then
		damInfo = UpdateDamageInfoAndApplyCritEffects(atTarget, rangedAttack, damInfo)
	end
	--DebugMessage("Daminfo Class:" .. tostring(damInfo.Class))
	--DebugTable(damInfo)
	damInfo.Class = GetDamageClassOverride(specOp, damInfo.Class)
	--DebugMessage("DAMCLASS " .. tostring(damInfo.Class))

	damInfo.isCrit = isCrit
	damInfo.damageAmt = damInfo.damageAmt + selfFXBonus
	if(facingMod > 1) then
		damInfo.IncreasedDamage = true
	end
	
	--MOD CODE HERE
	local myWep = mCombatInfo[wepHand].object
	if (myWep) then
		myWep:SendMessage("WeaponHitMessage", "hit", this, atTarget, damInfo.isCrit, damInfo.damageAmt)
	end
	--MODE CODE HERE

	--DebugMessage(tostring(this) .. "----------DAMAGE--------")
	--DebugTable(damInfo)
	--damInfo = CheckSpecialOverrides(atTarget,specOp, damInfo, isCrit)
	ApplyDamageToTarget(atTarget, damInfo)
end