require 'incl_dialogwindow'
require 'incl_guild_helpers'
require 'incl_Guild'

function InitGuildSystem()
		DebugMessage("Guild System: Init done!")

		RegisterCommand{ Command="g", AccessLevel = AccessLevel.Mortal, Func=Guild.SendMessage, Desc="Sends a guild message" }
		RegisterCommand{ Command="guild", AccessLevel = AccessLevel.Mortal, Func=Guild.GuildMenu, Desc="Opens the guild menu" }
		RegisterCommand{ Command="renameguild", AccessLevel = AccessLevel.Mortal, Func=Guild.Rename, Desc="Renames your guild (only 3 times!)" }

end

--RegisterEventHandler(EventType.ModuleAttached,"Guild",InitGuildSystem)

-- This function handles the player targeted event to check if all conditions are met
-- After that Guild.Invite is called to do the management stuff
RegisterEventHandler(EventType.ClientTargetGameObjResponse, "guildinvite",
	function(target,user)
		if( target == nil ) then
			return
		end

		if not( target:IsPlayer() ) then
			user:SystemMessage("Error: You can only invite players to join your guild")
			return
		end

		if (target == user) then
			user:SystemMessage("You cannot add yourself to your guild")
			return
		end

		local g = Guild.Get( user );
		local mg = Guild.Get( target );

			if ( g ~= nil and not(Guild.CheckAccessLevel(this, "Emissary"))) then
				user:SystemMessage("You need a higher rank to add a new member")
			return
			end

			if ( mg ~= nil and mg == g ) then
				user:SystemMessage("This person is already in your guild!")
				return
			end

			if ( mg ~= nil ) then
				user:SystemMessage("This person is already in a guild!")
				return
			end

			local guildInvite = target:GetObjVar("GuildInvitation")

			if (guildInvite ~= nil) then
				user:SystemMessage("This person is already considering another guild invitation")
				return
			end

			Guild.Invite(user, target)
			InviteGump(target, g.Name)
	end)

-- This Event handles the response of the guild window
RegisterEventHandler(EventType.DynamicWindowResponse, "GuildWindow",
function (user,buttonId)
	--DebugMessage("Guild: Handle Guild Windows Response");
	if (user == nil) then return end

	local g = Guild.Get(user)

	-- Handles the invite command of the dynamic window
	if (buttonId == "Invite") then

		if (g == nil) then
			user:SystemMessage("You are not in a guild");
			return
		end

		Guild.InviteTarget()
	end

	-- Handles the leave guild event
	if (buttonId == "Leave") then
		if (g == nil) then
			user:SystemMessage("You are not in a guild");
			return
		end

		if (this == g.Leader) then
			Guild.Disband(this)
		else
			Guild.Remove(user, this)
		end
		this:SendMessage("UpdateName")
	end

	if (buttonId =="Refresh") then
		Guild.GuildMenu()
	end

	if (buttonId =="Create") then
		local g = Guild.Get(this)

		if (g ~= nil) then
			this:SystemMessage("You are already in a guild")
			return
		end

		local guildInvite = this:GetObjVar("GuildInvitation")

		if (guildInvite ~= nil) then
			user:SystemMessage("You are already considering joining another guild")
			return
		end

		Guild.Create(this)
		this:SendMessage("UpdateName")
	end

	if (buttonId =="Rename") then
		this:SendClientMessage("EnterChat","/renameguild")
	end

	-- Handles the Edit player event (<) buttonId. The prefix
	-- Edit- is used to identify the command. That part is catched by
	-- the match() function. The rest is taken as function parameter gameobj number
	-- and converted back to a game obj
	if (buttonId:match("Edit-") ) then
		if (g == nil) then
			return
		end

		local objVarName = buttonId:sub(6)
		local varnum = tonumber(objVarName)

		DebugMessage("Guild: Edit request for "..tostring(varnum))

		Guild.EditMember(varnum, g)
	end

	--this:OpenDynamicWindow(newWindow)

end)

RegisterEventHandler(EventType.DynamicWindowResponse, "AcceptGuildInvitation",
function (user,buttonId)
	--DebugMessage("Guild: Handle Guild Invite Response");
	if (user == nil) then return end

	if (buttonId == "") then return end

	-- Handles the invite command of the dynamic window
	if (buttonId == "Accept") then
		Guild.Accept()
		this:SendMessage("UpdateName")
	end

	Guild.Decline()
end)

RegisterEventHandler(EventType.DynamicWindowResponse, "GuildMemberWindow",
function (user,buttonId)
	DebugMessage("GuildMember: Handle Guild Windows Response: "..buttonId);
	if (user == nil) then return end

	local g = Guild.Get(user)

	if (buttonId:match("Promote")) then
		local objVarName = buttonId:sub(8)
		local varnum = tonumber(objVarName)

		Guild.PromoteMember(varnum,g)
	end

	if (buttonId:match("Demote")) then
		local objVarName = buttonId:sub(7)
		local varnum = tonumber(objVarName)

		Guild.DemoteMember(varnum, g)
	end

	if (buttonId:match("Kick")) then
		local objVarName = buttonId:sub(5)
		local varnum = tonumber(objVarName)

		local mobile = GameObj(varnum)

		if (mobile ~= nil) then
			Guild.Remove(mobile, g)
			return
		end


		local mi = g.Members[varnum];

		if ( mi ~= nil ) then

			mi.Kick = true
			GetGuildStone():SetObjVar(g.Id, g)
		end
	end

end)

function InviteGump(mobile, guildname)
	local newWindow = DynamicWindow("AcceptGuildInvitation","Guild Invitation: "..guildname,300,150)

	newWindow:AddButton(20, 380, "Accept", "Accept", 0, 0)
	newWindow:AddButton(100, 380, "Decline", "Decline", 0, 0)

	mobile:OpenDynamicWindow(newWindow)
end

RegisterEventHandler(EventType.Message,"LoggedIn",function ()
	DebugMessage("Checking for dormant invitations")

	local guildInvite = this:GetObjVar("GuildInvitation")

	if (guildInvite ~= nil) then

		local guild = Guild.ValidateInvitation(guildInvite)

		if(guild ~= nil) then
			InviteGump(this, guild.Name)
		else
			this:SetObjVar("GuildInvitation", nil)
		end

		return
	end

	local guild = Guild.Get(this)

	Guild.UpdateMemberInfo(this,guild)

	if (guild ~= nil) then
		Guild.SendToAll( this, guild, "(Has logged on)");
	end
end)
