-- NPC Editor

require "incl_mobile_helpers"
require "editor.template_dump"

TEMPLATE_PATH = "mods\\arcanima\\templates\\ac_mob_ed\\"
BASE_TEMPLATE_PATH = "mods\\arcanima\\templates\\ac_mob_templates\\"

-- Copied from vanilla, be sure to update when needed
BodyTypesMale = {
	{Name = "Male A",},
	{Name = "Male B",Template = "head_male02"},
	{Name = "Male C",Template = "head_male03"},
	{Name = "Male D",Template = "head_male04"},
	{Name = "Male E",Template = "head_male05"},
}
BodyTypesFemale = {
	{Name = "Female A",Template = "head_female01"},
	{Name = "Female B",Template = "head_female02"},
	{Name = "Female C",Template = "head_female03"},
	{Name = "Female D",Template = "head_female04"},
	{Name = "Female E",Template = "head_female05"},
}

HairTypesMale = {
	{Name = "Regular",Template = "hair_male"},
	{Name = "Bangs",Template = "hair_male_bangs"},
	{Name = "Buzzcut",Template = "hair_male_buzzcut"},
	{Name = "Bald"},
	{Name = "Messy",Template = "hair_male_messy"},
	{Name = "Nobleman",Template = "hair_male_nobleman"},
	{Name = "Roguish",Template = "hair_male_rougish"},
	{Name = "Side-Undercut",Template = "hair_male_sideundercut"},
	{Name = "Undercut",Template = "hair_male_undercut"},
	{Name = "Windswept",Template = "hair_male_windswept"},
}

HairTypesFemale = {
	{Name = "Pony-Tail A",Template = "hair_female"},
	{Name = "Bob",Template = "hair_female_bob"},
	{Name = "Bun",Template = "hair_female_bun"},
	{Name = "Buzzcut",Template = "hair_female_buzzcut"},
	{Name = "Pigtails",Template = "hair_female_pigtails"},
	{Name = "Pony-Tail B",Template = "hair_female_ponytail"},
	{Name = "Shaggy",Template = "hair_female_shaggy"},
}

HairColorTypes = {
	{Name = "Dark Brown", Color = "0xFF554838"},
	{Name = "Brown", Color = "0xFFA56B46"},
	{Name = "Light-Brown", Color = "0xFFA7856A"},
	{Name = "Red", Color = "0xFFB55239"},
	{Name = "Black", Color = "0xFF191919"},
	{Name = "Gray", Color = "0xFFB7A69E"},
	{Name = "Blonde", Color = "0xFFB9B184"},
	{Name = "Dark Red", Color = "0xFF380000"},
	{Name = "Orange", Color = "0xFFFF7F00"},
	{Name = "White", Color = "0xFFFFFFFF"},
	{Name = "Purple", Color = "0xFF990099"},
	{Name = "Dark Blue", Color = "0xFF012265"},
	{Name = "Green", Color = "0xFF339933"},
	{Name = "Blue", Color = "0xFF3399FF"},
	{Name = "Pink", Color = "0xFFFF66A3"},
}

CreationTemplates = {
	{name = "Human Male", template = "editor_template_male"},
	{name = "Human Female", template = "editor_template_female"}
}

AIModulePresets = {
	{Name = "Basic NPC", 			  "base_ai_npc", "ai_modules.dialogue"},
	{Name = "Basic NPC + Reputation", "base_ai_npc", "ai_modules.dialogue", "relation_system",  "ai_modules.repution_aggro"},
	{Name = "Healer", 				  "base_ai_npc", "ai_modules.dialogue", "ai_modules.healer"}
}

-- Gender specific tables, see ParseGenderTables()
GenderBodyTable = nil
GenderHairTable = nil

-- Body part indexes in tables above, see ParseBodyIndexes()
BodyIndex = 1
HairIndex = 1

CreationIndex = 1
CreationTemplate = CreationTemplates[1].template
CreationLoc = nil
CreationTries = 10

npc = nil

NavBarItems = {
	{name = "Appearance", 	display = "Appearance", 	size = 160},
	{name = "Behavior", 	display = "Behavior", 		size = 160},
	{name = "SysVars", 		display = "System ObjVars", size = 160},
}
EditTabs = {}

function RequestCreateNewNpc(index)
	CreationIndex = index
	CreationTemplate = CreationTemplates[index]

	local width = 480
	local height = 295

	local window = DynamicWindow("npcCreator", "Create NPC", width, height, 600, 400, "SemiTransparent")

	window:AddLabel(100, 20, "Select Template", 0, 0, 23, "center")
	for i,v in ipairs(CreationTemplates) do
		local state = ""
		if i == index then
			state = "pressed"
		end

		window:AddButton(40, 26 + 23 * i, "Select", "", 0, 19, "", "selectCreationTemplate "..i, false, "Selection", state)
		window:AddLabel( 64, 26 + 23 * i, " "..v.name, 0, 0, 20)
	end

	window:AddTextField(35, 200, 260, "template")
	window:AddButton(310, 10, "CreateNpc", "Create", 125, 220, "", "", false,"TabButton")

	this:OpenDynamicWindow(window, this)
end

function CreateAndEdit(success, loc)
	if not success then return end

	this:CloseDynamicWindow("npcCreator")
	CreationLoc = loc

	--DebugMessage(BASE_TEMPLATE_PATH..CreationTemplates[CreationIndex].template)
	local baseTemp = io.open(BASE_TEMPLATE_PATH..CreationTemplates[CreationIndex].template..".xml", "r")
	local str = baseTemp:read("*a")
	baseTemp:close()

	local name = GetTemplateName(CreationTemplate)
	local newTemp = io.open(TEMPLATE_PATH..name..".xml", "w")
	newTemp:write(str)
	newTemp:close()

	ReloadTemplates()
	this:ScheduleTimerDelay(TimeSpan.FromSeconds(4), "wait_template_reload") -- TODO 4 seconds is too much, probably...
end

RegisterEventHandler(EventType.ClientTargetLocResponse, "editor_CreateNPC", CreateAndEdit)

RegisterEventHandler(EventType.Timer, "wait_template_reload", function ()
		CreateObj(GetTemplateName(CreationTemplate), CreationLoc, "create_and_edit")
	end)

RegisterEventHandler(EventType.CreatedObject, "create_and_edit", function (success, gameObj)
		if not success then
			if CreationTries < 11 then
				CreationTries = CreationTries - 1
				this:ScheduleTimerDelay(TimeSpan.FromSeconds(1), "wait_template_reload")
			else
				this:SystemMessage("[FF0000]Failed to create NPC! Please check console.")
			end
			return
		end

		npc = gameObj
		ParseGenderTables()
		ParseBodyIndexes()
		OpenEditor("Appearance", 1)
	end)

RegisterEventHandler(EventType.DynamicWindowResponse, "npcCreator", function (user, buttonId, fields)
		if buttonId == "CreateNpc" then
			this:RequestClientTargetLoc(this, "editor_CreateNPC")
			CreationTemplate = fields.template
		end
	end)

RegisterEventHandler(EventType.ClientUserCommand, "selectCreationTemplate", function (index)
		RequestCreateNewNpc(tonumber(index))
	end)

function RequestEditNpc()
	this:RequestClientTargetGameObj(this, "editor_TargetNPC")
end

function OpenEditor(tab, delay)
	if delay ~= nil then
		RegisterSingleEventHandler(EventType.Timer, "editor_open_delayed",
			function ()
				OpenEditor(tab) -- Upvalues are awesome
			end)

		return this:ScheduleTimerDelay(TimeSpan.FromSeconds(delay), "editor_open_delayed") -- Return did it work
	end

	local name = npc:GetName()
	tab = tab or "Appearance"

	local width = 800
	local height = 500

	local window = DynamicWindow("npcEditor", "Npc Editor - "..StripColorFromString(name), width, height, 0.3 * width, 0.3 * height, "SemiTransparent")

	EditTabs[tab](window)
	CreateNavBar(window, 11, 1)
	CreateSaveTemplate(window, 400, 390)
	this:OpenDynamicWindow(window, this)

	return true -- It worked
end

-- Buttons, like save, update and so on
function CreateNavBar(window, x, y)
	local usedWidth = x
	for priority,data in ipairs(NavBarItems) do
		window:AddButton(usedWidth, y, data.name, data.display, data.size, 0, data.tooltip or "", data.serverCmd or "", false,"TabButton")
		usedWidth = usedWidth + data.size
	end
end

function CreateSaveTemplate(window, x, y)
	window:AddTextField(x, y + 6, 190, "newTemplate")
	window:AddButton(x + 200, y, "Save", "Save Template", 0, 0, "", "", false)
end

function EditTabs.Appearance(window)
	CreateEquipSlots(window, 230, 50)
	CreatePortrait(window, 50, 70)
	CreateTopInfo(window, 50, 400)
	CreateBodyParts(window, 400, 250)
	CreateNameChange(window, 400, 50)
	CreateMobileType(window, 400, 150)
end

-- Creates the portrait of NPC
function CreatePortrait(window, x, y)
	window:AddPortrait(x, y, 180, 270, npc, "full")
end

-- Creates equip slots
function CreateEquipSlots(window, x, y)
	local sprite = "EquipSlotBG"
	local slots = {"Head", "Chest", "Legs", "RightHand", "LeftHand"}

	for i,slot in pairs(slots) do
		window:AddImage(x, y + i * 50, sprite, 32, 32, "Simple")
		local action = {
			ID = slot,
			ActionType = "EquipSlot",
			TargetObject = npc,
			DisplayName = "", -- Huh? In vanilla this is displayName, which is clearly a nil global
			Enabled = true,
			ServerCommand = "equipSlotClicked " .. slot
		}
		window:AddUserAction(x, y + i * 50, action, 28, 28) -- You can't actually pick up there, maybe vanilla bug?
	end
end

-- Top info (template name, etc.)
function CreateTopInfo(window, x, y)
	local template = "Template: "..npc:GetCreationTemplateId()
	window:AddLabel(x, y, template, 0, 0, 20)
end

-- Body editor
function CreateBodyParts(window, x, y)

	window:AddImage(x, y, "FullBorder", 250, 55, "Sliced")

	-- Change Body Buttons
	window:AddLabel(x + 125, y +  5, "[F3F781]Body Type[-]", 250, 0, 23, "center")
	window:AddLabel(x + 125, y + 28, GenderBodyTable[BodyIndex].Name, 250, 0, 18, "center")
		-- Bigger functional area
	window:AddButton(x +   3, y + 4, "BodyLeft" , "", 123, 47, "", "", false, "Invisible")--"LeftArrow")
	window:AddButton(x + 126, y + 4, "BodyRight", "", 122, 47, "", "", false, "Invisible")--"RightArrow")
		-- Visible Buttons
	window:AddButton(x +   7, y + 19, "BodyLeft" , "", 123, 47, "", "", false, "LeftArrow")
	window:AddButton(x + 235, y + 19, "BodyRight", "", 125, 47, "", "", false,  "RightArrow")

	--Change Hair Buttons
	y = y + 57
	window:AddImage(x, y, "FullBorder", 250, 55, "Sliced")
	window:AddLabel(x + 125, y +  5, "[F3F781]Hair[-]", 250, 0, 23,"center")
	window:AddLabel(x + 125, y + 28, GenderHairTable[HairIndex].Name, 250, 0, 18, "center")
		-- Bigger functional area
	window:AddButton(x +   3, y + 4, "HairLeft" , "", 123, 47, "", "", false, "Invisible")
	window:AddButton(x + 126, y + 4, "HairRight", "", 123, 47, "", "", false, "Invisible")
		-- Visible Buttons
	window:AddButton(x +   7, y + 19, "HairLeft", "", 0, 0, "", "", false, "LeftArrow")
	window:AddButton(x + 235, y + 19, "HairRight", "", 0, 0, "", "", false, "RightArrow")
end

-- Creates separate gender tables according to NPC's gender
function ParseGenderTables()
	if npc:IsMale() then
		GenderBodyTable = BodyTypesMale
		GenderHairTable = HairTypesMale
	elseif npc:IsFemale() then
		GenderBodyTable = BodyTypesFemale
		GenderHairTable = HairTypesFemale
	end
end

-- Parses body indexes for new NPC
function ParseBodyIndexes()
	local head = npc:GetEquippedObject("BodyPartHead")
	local headTemplate
	if head ~= nil then headTemplate = head:GetCreationTemplateId() end
	for i,v in ipairs(GenderBodyTable) do
		if headTemplate == v.Template then
			BodyIndex = i
			break
		end
	end

	local hair = npc:GetEquippedObject("BodyPartHair")
	local hairTemplate
	if hair ~= nil then hairTemplate = hair:GetCreationTemplateId() end
	for i,v in ipairs(GenderHairTable) do
		if hairTemplate == v.Template then
			HairIndex = i
			break
		end
	end
end

-- Replaces body part with new one
function ReplaceBodyPart(part, newTemplate)
	--DebugMessage("Replacing "..part.." with "..newTemplate)
	local current = npc:GetEquippedObject(part)
	if current ~= nil then
		current:Destroy()
	end
	if newTemplate == nil then return end
	local ident = "created_object"
	if part == "BodyPartHead" then
		ident = "created_body_type"
	elseif part == "BodyPartHair" then
		ident = "created_hair"
	end
	CreateEquippedObj(newTemplate, npc, ident)
end

-- Validates table index - specially for body part tables
function ValidateIndex(table, index)
	if index > #table then index = 1 end
	if index < 1 then index = #table end
	return index
end

-- Changes given bodypart by index in table
function EquipByIndex(part, index)
	--DebugMessage("Equip "..part.." by index "..index)
	local template
	if part == "BodyPartHead" then
		template = GenderBodyTable[index].Template
	elseif part == "BodyPartHair" then
		template = GenderHairTable[index].Template
	end

	ReplaceBodyPart(part, template)
end

function CreateNameChange(window, x, y)
	window:AddLabel(x + 5, y, "Change Name", 0, 0, 23)
	window:AddTextField(x, y + 30, 250, "newName")
	window:AddButton(x + 255, y + 26, "ChangeName", "Change", 0, 25, "", "", false)
end

function CreateMobileType(window, x, y)
	local mobType = npc:GetMobileType()
	if mobType == "Animal" then
		mobType = "[FFFF00]Animal"
	elseif mobType == "Monster" then
		mobType = "[FF0000]Monster"
	elseif mobType == "Friendly" then
		mobType = "[00FF00]Friendly"
	end

	window:AddLabel(x, y, "Type: "..mobType, 0, 0, 23)
	window:AddTextField(x , y + 30, 100, "mobType")
	window:AddButton(x + 105, y + 26, "ChangeMobType", "Change", 0, 25, "", "", false)
	window:AddLabel(x, y + 60, "(Animal, Monster, Friendly or Player)", 0, 0, 23)
end

function EditTabs.Behavior(window)
	CreateStats(window, 425, 120)
	CreateAIModules(window, 15, 80)
	CreateToggleFreeze(window, 410, 45)
	CreatePresetButton(window, 510, 45)
end

-- AI module viewer
function CreateAIModules(window, x, y)
	local scroll = ScrollWindow(x, y, 380, 270, 25)

	local mods = npc:GetAIModules()
	for i,mod in ipairs(mods) do
		local element = ScrollElement()
		element:AddLabel(10, 5, mod, 0, 0, 18)
		local serverCmd = "editAIModules detach "..mod
		element:AddButton(275, 0, "DetachAIModule", "Detach", 60, 25, "", serverCmd, false, "SimpleButton")
		--element:AddImage(0, 0, "FullBorder", 360, 30, "Sliced")
		element:AddButton(215, 0, "ConfigAIModule", "Config", 60, 25, "", "editAIModules config "..mod, false, "SimpleButton")

		scroll:Add(element)
	end

	window:AddLabel(x + 190, y - 30, "AI Modules", 380, 0, 23, "center")
	window:AddScrollWindow(scroll)
	--window:AddImage(x, y - 7, "FullBorder", 420, 280, "Sliced")

	window:AddTextField(x, y + 285, 280, "newAIModule")
	window:AddButton(x + 290, y + 285, "AttachAIModule", "Attach", 80, 25, "", "", false, "SimpleButton")
end

-- Freeze/unfreeze NPC
function CreateToggleFreeze(window, x, y)
	local frozen = npc:IsMobileFrozen(true)
	local text = "ERROR!"
	if frozen then
		text = "Unfreeze"
	else
		text = "Freeze"
	end

	window:AddButton(x, y, "ToggleFreeze", text, 100, 0, "", "", false, "SimpleButton")
end

-- Stats (str, agi, int)
function CreateStats(window, x, y)
	window:AddLabel(x + 35, y - 30, "Stats", 0, 0, 23)

	window:AddLabel(x, y, "Str: "..npc:GetStr(), 0, 0, 23)
	window:AddLabel(x, y + 30, "Agi: "..npc:GetAgi(), 0, 0, 23)
	window:AddLabel(x, y + 60, "Int: "..npc:GetInt(), 0, 0, 23)

	window:AddTextField(x + 80, y, 30, "newStr")
	window:AddTextField(x + 80, y + 30, 30, "newAgi")
	window:AddTextField(x + 80, y + 60, 30, "newInt")

	window:AddButton(x + 35, y + 100, "UpdateStats", "Update", 0, 0, "", "", false)
end

function UpdateStats(str, agi, int)
	str = str or ""
	agi = agi or ""
	int = int or ""

	if str ~= "" then
		local num = tonumber(str)
		if num ~= nil and num >= 0 then
			npc:SetStr(num)
		end
	end

	if agi ~= "" then
		local num = tonumber(agi)
		if num ~= nil and num >= 0 then
			npc:SetAgi(num)
		end
	end

	if int ~= "" then
		local num = tonumber(int)
		if num ~= nil and num >= 0 then
			npc:SetInt(num)
		end
	end
end

-- Presets for AI modules
function CreatePresetButton(window, x, y)
	window:AddButton(x, y, "AIPresets", "Presets", 100, 0, "", "", false, "SimpleButton")
end

function OpenAIPresetWindow()
	local window = DynamicWindow("aiPresets", "AI Presets", 500, 300, 600, 400, "SemiTransparent")
	local y = 5
	local x = 15
	for i,preset in ipairs(AIModulePresets) do
		window:AddButton(x, y + (i-1) * 35, "SetFromPreset", preset.Name, 200, 0, "", "setFromPreset "..i, true, "TabButton")
	end

	this:OpenDynamicWindow(window, this)
end

RegisterEventHandler(EventType.ClientUserCommand, "setFromPreset", function (index)
	local preset = AIModulePresets[tonumber(index)]
	if preset == nil then
		this:SystemMessage("Error: AI preset not found!")
		return
	end

	local mods = {}
	for i,name in ipairs(preset) do
		mods[i] = name
	end
	npc:SetObjVar("AIModules", mods)
	ReloadModule("ai_modular")

	OpenEditor("Behavior", 0.5)
end)

function HandleData(user, buttonId, fields)
	if buttonId == "Save" then -- Dump template!
		Save(fields.newTemplate)
	elseif buttonId == "BodyLeft" then -- Check for body part changes
		BodyIndex = ValidateIndex(GenderBodyTable, BodyIndex - 1)
		EquipByIndex("BodyPartHead", BodyIndex)
		OpenEditor()
	elseif buttonId == "BodyRight" then
		BodyIndex = ValidateIndex(GenderBodyTable, BodyIndex + 1)
		EquipByIndex("BodyPartHead", BodyIndex)
		OpenEditor()
	elseif buttonId == "HairLeft" then
		HairIndex = ValidateIndex(GenderHairTable, HairIndex - 1)
		EquipByIndex("BodyPartHair", HairIndex)
		OpenEditor()
	elseif buttonId == "HairRight" then
		HairIndex = ValidateIndex(GenderHairTable, HairIndex + 1)
		EquipByIndex("BodyPartHair", HairIndex)
		OpenEditor()
	elseif buttonId == "AttachAIModule" then -- Attach that AI module
		if fields.newAIModule ~= nil and fields.newAIModule ~= "" then
			npc:AddAIModule(fields.newAIModule)
			OpenEditor("Behavior", 0.3)
		end
	elseif buttonId == "ToggleFreeze" then -- Freeze or unfreeze
		local frozen = npc:IsMobileFrozen(true) -- We check it here instead of button creation so bugs are not so dangerous
		if frozen then
			npc:SetMobileFrozen(false, false)
		else
			npc:SetMobileFrozen(true, true)
		end
		OpenEditor("Behavior", 0.1)
	elseif buttonId == "UpdateStats" then -- Update stats which have changed values
		UpdateStats(fields.newStr, fields.newAgi, fields.newInt)
		OpenEditor("Behavior", 0.5)
	elseif buttonId == "ChangeName" then -- Change display name
		local newName = fields.newName
		if newName ~= nil and newName ~= "" then
			npc:SetName(newName)
			OpenEditor("Appearance", 0.1)
		end
	elseif buttonId == "ChangeMobType" then
		local newType = fields.mobType
		if newType ~= nil and newType ~= "" then
			if newType ~= "Animal" and newType ~= "Monster" and newType ~= "Friendly" and newType ~= "Player" then
				this:SystemMessage("Unknown mobile type "..newType..", using Animal instead.")
				newType = "Animal"
			end

			npc:SetMobileType(newType)
			OpenEditor("Appearance", 0.1)
		end
	elseif buttonId == "AIPresets" then
		OpenAIPresetWindow()
	else -- If nothing else, this in toolbar action... Hopefully
		OpenEditor(buttonId)
	end
end

-- Prefix template name as needed
function GetTemplateName(name)
	if name:sub(1, 4) ~= "edt_" then -- TODO maybe remove when editor is stable!
		name = "edt_"..name
	end

	return name
end

-- Dump to template
function Save(newName, npcObj)
	if newName == "" or newName == nil then
		newName = npc:GetCreationTemplateId()
	end
	npcObj = npcObj or npc

	newName = GetTemplateName(newName)

	DumpToTemplate(npcObj, TEMPLATE_PATH..newName..".xml", true)
	this:SystemMessage("Saved NPC template "..newName)
end

AddToolbarItem("editNpc", "Edit NPC", RequestEditNpc)
AddToolbarItem("createNpc", "Create NPC", function ()
	RequestCreateNewNpc(1)
end)

RegisterEventHandler(EventType.ClientUserCommand, "equipSlotClicked", -- Copied from vanilla base_player_charwindow
	function(equipSlot)
		if(not(IsInTableArray (ITEM_SLOTS,equipSlot))) then
			DebugMessage("ERROR: Client specified invalid equipment slot! "..equipSlot)
			return
		end

		local carriedObject = this:CarriedObject()
		if(carriedObject ~= nil) then
			local carriedEquipSlot = GetEquipSlot(carriedObject)
			if(carriedEquipSlot ~= equipSlot) then
				this:SystemMessage("You can not equip that there.")
			else
				DoEquip(carriedObject, npc)
			end
		else
			local equippedObj = this:GetEquippedObject(equipSlot)
			if(equippedObj ~= nil) then
				this:SendMessage("PickupObject",equippedObj)
			end
		end
	end)

RegisterEventHandler(EventType.ClientUserCommand, "editAIModules", function (action, mod)
		if action == "detach" then
			npc:DelAIModule(mod)
			OpenEditor("Behavior", 0.3)
		elseif action == "attach" then -- Unused for now
			npc:AddAIModule(mod)
			OpenEditor("Behavior", 0.3)
		elseif action == "config" then
			this:SendMessage("Config not yet implemented!")
		end
	end)

RegisterEventHandler(EventType.ClientTargetGameObjResponse, "editor_TargetNPC", function (target, targeter)
		npc = target
		ParseGenderTables()
		ParseBodyIndexes()
		OpenEditor()
	end)

RegisterEventHandler(EventType.DynamicWindowResponse, "npcEditor", HandleData)
