--- Main Cronjob Handling Function, called by timer
-- @param CallerData = {string} = {caller} name (for debug purposes)
-- The handler takes the 1st entry of SCHEDULE and processes it
-- It is the Handler of the "cron" timer
function HandleCron(CallerData)
	-- Stop timer if any are left for some odd reason (asynchronous bitches)
		if this:HasTimer("Cron") then 
			this:RemoveTimer("Cron") 
			DebugMessage("Stopping Timer 'Cron'")
		end

	-- DEBUG: Who has summoned me?
		DebugMessage("Entering HandleCron from <= "..tostring(CallerData[1]).." )", true)

	-- Nothing to schedule, so we go home drink .coffee
		if ( ( SCHEDULE == {}) or (SCHEDULE == nil or SCHEDULE[#SCHEDULE])==nil ) then 
			DebugMessage("No SCHEDULE - leaving", true)
			return
		end

	-- stuff we need
	local entry   = (SCHEDULE[1])[2]  -- the cronentry as stored in the schedule table
	local curRun = (SCHEDULE[1])[1]  	--- scheduled time for the curren run in game minutes since game midnight
	local myObj   = GameObj(entry.Id) -- the subscribing object
	
	
	----------------------------
	-- detect a possible timeleap, e.g. after a /settime by an admin
	-- if someone used /settime we have a problem we must rebuild the SCHEDULE and reschedule
		local timeleap = math.abs(GetGMinutesSinceGMidnight()-curRun)
		
		if (timeleap>TIMELEAP_TOLERANCE) then 
			DebugMessage("Cron: Timeleap detected("..timeleap.."): Rebuilding Schedule and eventually dropping scheduled events ...")
			-- We can just use the midnight routine for this and then leave
			-- This will also readjust the midnight timer
				-- this:FireTimer("Midnight","Timeleap Detector")
				HandleMidnight("Timeleap Detector")
			return 
		end

	
 	
	---------------------------------------------------------------------
	--- Scheduling our next event:
	---------------------------------------------------------------------
 	
 	--- Construct the later message string which gets sent to the subscribing GameObj:
 		local MessageData={"Cron" .. entry.Id .."_".. entry.Name,entry.Data}

 	
 	-- schedule next event and determine the next runtime for this entry
		-- the +1 is necessary to prevent infinite loops from overdue events.
		local firenext=GetNextPingtime(GetGMinutesSinceGMidnight()+1, entry.StartTime[4]*60+entry.StartTime[5], entry.Interval)

		-- only if there is no midnight wraparound/overflow we add the entry
			if (firenext < 1440) then
				-- Get length of SCHEDULE table
				local next=#SCHEDULE+1
				-- append our entry to the SCHEDULE
				SCHEDULE[next]={firenext, entry}
				-- for debugging purposes - not really needed for cron to function
				this:SetObjVar("schedule", SCHEDULE)
			-- if there is a wraparound we drop the entry 
			else
				DebugMessage("Cron Main loop: Scheduled event dropped (Reason: Midnight Warp) at minute: ".. firenext)
			end

		--remove first entry (the last run we will just fire - we already have the SenMessage Data (see above))
		table.remove(SCHEDULE, 1)
		-- sort remaining table
		table.sort(SCHEDULE, CompareSchedulerEntries)

		-------------------------------
		-- Seems we're ready to shoot the event
 			myObj:SendMessage(table.unpack(MessageData))


		-- calculate milliseconds from now to Pingtime for the first entry of the SCHEDULE
		-- using the time factor to real time
	  if  ( SCHEDULE[1] ~= {}) and (SCHEDULE[1] ~= nil ) then 
		  local nextrun=(SCHEDULE[1])[1]  --- next run in minutes since midnight
			local msecs=( nextrun-GetGMinutesSinceGMidnight())*60*1000 * FACTOR
			-- check for overdue events and fire them immediately (1 millisecond delay)
			if msecs<=0 then msecs = 1 end
			---------------------
	 		-- (re)start the "cron" timer 
			this:ScheduleTimerDelay(TimeSpan.FromMilliseconds(msecs), "Cron", {"CronHandler-Regular"} ) 
			-- DebugMessage:
			DebugMessage('Cron Main Loop: Pinging in '.. math.floor(msecs/1000) .. " seconds") 
		end
	end


--- Comparison function for sorting a table of tables. 
-- First entry in subtable defines order. 
-- Lowest to highest
function CompareSchedulerEntries(entry1, entry2)
	--LuaDebugCallStack("CompareSchedulerEntries")
	-- lower value comes first
	if entry1[1]<entry2[1] then return true end
	return false
	end

function HandleCronSubscribe(cronentry)
	-- add the entry delivered by the subscribing object into the crontab
	crontab[cronentry.Id .. "_" .. cronentry.Name]=cronentry
	--making it persistent
	this:SetObjVar("crontab", crontab)
	--debug
	DebugMessage("Cron event: Received subscription " .. cronentry.Id .. "_" .. cronentry.Name)
	RebuildSchedule(crontab, true)
	end

function HandleCronUnSubscribe(cronentry)

	crontab[tostring(cronentry.Id) .. "_" .. cronentry.Name]=nil
	-- making it persistent
	this:SetObjVar("crontab", crontab)
	-- rebuild the Schedule
	RebuildSchedule(crontab, true)
	--debug
	DebugMessage("Cron event: Bye bye Entry: " .. cronentry.Id .. "_" .. cronentry.Name, true)
	end

function HandleCronRemoveObject(Id)  
	-- remove all crontab entries for the object with the handle "Id"
	-- has to be requested with  SendMessage("CronRemoveObject", this.Id) by the object
	for key,value in pairs(crontab) do 
		if ( string.sub( key, 1, #(Id.."_") ) == (Id .. "_") ) then 
			-- remove entry entirely
			crontab[key]=nil
			--debug
			DebugMessage("Cron event: Removed Key: " .. key, true)	
		end
	end
	-- making it persistent
	this:SetObjVar("crontab", crontab)
	-- Rebuild the SCHEDULE
	RebuildSchedule(crontab,true)
	--debug
	DebugMessage("Cron event: Bye bye Object: " .. Id, true)	

	end

function HandleCronRebuildSchedule()
	--LuaDebugCallStack("HandleCronRebuildSchedule")
 	RebuildSchedule(crontab,true)
 end 

-- Housekeeping:  Remove invalid entries from lost objects
function CronCheckEntries()
	local myObj=nil
	for key, entry in pairs(crontab) do
		myObj=GameObj(tonumber(entry.Id))
		if (myObj == nil ) or not (myObj:IsValid())then
			--debug
			DebugMessage("CronCheckEntries(): removing Id:  "..entry.Id, true)
			crontab[key]=nil
		end
	end
	-- make changes persistent
	this:SetObjVar("crontab", crontab)
	-- rebuild SCHEDULE
	RebuildSchedule(crontab,true)
 end
-------------------------------------------------------------------------------------
--- This function rebuilds the schedule of events from the crontab
-- this has to happen on startup for example
-- the schedule itself is not persistent over server restarts, 
-- only the crontab is stored as an objvar on the cron object.
-- bIgnoreStartTime is a bool which fixes schedule entries over 1440 (=more than 24 hours)
-------------------------------------------------------------------------------------
function RebuildSchedule(crontab, bIgnoreStartTime)
	--LuaDebugCallStack("RebuildSchedule")
	--------------------
	-- preliminary stuff
	--------------------
		-- Stop timer if any left
		if this:HasTimer("Cron") then 
			this:RemoveTimer("Cron") 
			DebugMessage("Stopping Timer 'Cron'",true)
		end


	-- Rebuild the schedule table
		Reschedule()
		--For debugging make the new SCHEDULE perma and inspectable
		this:SetObjVar("schedule",SCHEDULE)	

	-- CHECK: return on invalid SCHEDULE
		if (( SCHEDULE==nil) or (SCHEDULE=={}) or (SCHEDULE[#SCHEDULE]==nil) )then 
			DebugMessage("RebuildSchedule: Nothing to schedule - leaving", true) 
			return 
		end

	-- sort the table by scheduled times
		table.sort(SCHEDULE, CompareSchedulerEntries)

	-- Prepare the timer	
		-- calculate milliseconds from now  to Pingtime for the first entry
		-- calculate time with factor to real time Should be positive (check downwards)
		local msecs=( (SCHEDULE[1])[1]-GetGMinutesSinceGMidnight() )*60*1000 * FACTOR
		
	-- check for overdue events and fire them immediately (no delay)
		if msecs<0 then msecs = 0 end

	-- Now  (re)start the timer 
		DebugMessage("RebuildSchedule: Next scheduled event at game minute: "..tostring((SCHEDULE[#SCHEDULE])[1]),true)
		DebugMessage('RebuildSchedule: Calling Cron main loop in ==> '..math.floor(msecs/1000).." seconds", true) 
		this:ScheduleTimerDelay(TimeSpan.FromMilliseconds(msecs), "Cron", {"RebuildHandler"} ) 

 end 

function Reschedule()
	-- using the global SCHEDULE and crontab and current time
	local firenext, next
	-- flush SCHEDULE
	SCHEDULE={}
	-----------------------------------------------
	-- iterate over the whole crontab and rebuild
	-----------------------------------------------
		--debug
		--DebugMessage("Crontab not empty - rebuilding schedule entries ...", true) 

		for key, entry in pairs(crontab) do 
			-- calculate next time to ping this entry. Currently only evaluating hour and minute of starttime
			firenext=GetNextPingtime(GetGMinutesSinceGMidnight(), entry.StartTime[4]*60+entry.StartTime[5], entry.Interval)

			-- Must be between 0 and 1439 game time minutes. 24:00 = 00:00 = 0 and not 1440
			if (firenext < 1440) then --all is okay and ...
				-- put entry into the schedule at the end of it
				next=#SCHEDULE+1
				SCHEDULE[next]={firenext, entry}
			else
				-- Otherwise we drop the event until the midnight routine
				-- debug
				DebugMessage("Reschedule: Dropped an event at game minute "..firenext.." (overflow)")
			end
		end

	return
end




-------------------------------------------------------------------------------
--- Calculate the next ping depending on the current time, the starttime of the crontab entry and the Interval to ping
-- all parameter given in in game minutes since midnight
-- gametime = gametime when to restart the cron service
-- starttime = time when to start the pinging
-- interval = ping interval
-------------------------------------------------------------------------------
function GetNextPingtime(gametime, starttime, interval )
	--LuaDebugCallStack("GetNextPingtime")
	--if gametime<starttime then
		return starttime + (  math.ceil( (gametime-starttime)/interval )*interval  )
	--else
	--	return starttime + (  math.floor( (gametime-starttime)/interval )*interval  )
	--end
	end

function HandleMidnight(...)
	local args=table.pack(...)
	DebugMessage("Entering "..tostring(this:GetName())..":HandleMidnight() from "..tostring(args[1]) )

	-- kill any cron timer (main loop) if any exists
	if this:HasTimer("Cron") then this:RemoveTimer("Cron") end
	
	-- Kill and restart Midnight rebuild timer
	if this:HasTimer("Midnight") then this:RemoveTimer("Midnight") end
	local msecs= (24-GetGameTimeOfDay()) *3600*1000  * FACTOR
	this:ScheduleTimerDelay(TimeSpan.FromMilliseconds(msecs), "Midnight", "HandleMidnight" ) 
	
	DebugMessage("Midnight Handler: Scheduling next run in ".. math.floor(msecs/(100*60))/10 .." minutes.")

	--rebuilding SCHEDULE at midnight
	RebuildSchedule(crontab,true)

end

---------------------------------------------------------------------------------
--- Cron Startup Sequence
---------------------------------------------------------------------------------
function CronStartup()
	
	-- kill any cron timer (main loop) if any exists
		if this:HasTimer("Cron") then this:RemoveTimer("Cron") end
	--- DOC: FIXME
	--- SCHEDULE  
	-- The schedule table holds a list of scheduled events consisting of a time to fire and the crontab entry.
	-- tables with cron entries and scheduled times when to execute
	-- these tables are keyed with numbers corresponding to in-game minutes when to fire
	-- once the scheduler is queried all entries keyed accordingly get called and removed
	-- each called entry will place itself into the next appropriately keyed table
	-- an entry to be called at exactly 1:00h would be in a table indexed as 
	-- 60 = the amount of minutes after midnight
	SCHEDULE={}
	-- crontab workcopy
	crontab={}

	-- Set up persistent crontab on cron GameObject
	if (this:HasObjVar("crontab")) then 
		crontab = this:GetObjVar("crontab")
		CronCheckEntries() -- also start timer loop via Rebuild routine
		--debug
		--DebugMessage("Found crontab ObjVar on myself("..this.Id..") assigning to workcopy", true)
	else
		this:SetObjVar("crontab", crontab)
		DebugMessage("Found no crontab ObjVar on myself("..this.Id..") - creating an empty one", true)
	end

	-----------------------------------------

	-- EVENT HANDLERS

	RegisterEventHandler(EventType.Timer, "CronHousekeeping", 
		function ()
			DebugMessage("CRON "..this.Id..": HOUSEKEEPING START", true)
			CronCheckEntries()
			-- stop timer if it exists
		    if this:HasTimer("CronHousekeeping") then this:RemoveTimer("CronHousekeeping") end
			this:ScheduleTimerDelay(TimeSpan.FromMilliseconds(HOUSEKEEPING_INTERVAL), "CronHousekeeping", {} ) 
			DebugMessage("CRON "..this.Id..": HOUSEKEEPING FINISHED", true)
		end)
	if this:HasTimer("CronHousekeeping") then this:RemoveTimer("CronHousekeeping") end
	this:ScheduleTimerDelay(TimeSpan.FromMilliseconds(HOUSEKEEPING_INTERVAL), "CronHousekeeping", {} ) 


	-----------------------------------------
	----------------------------------------
	-- The main loop catching the timer
	----------------------------------------
	RegisterEventHandler(EventType.Timer, "Cron", HandleCron)

	------------------------------------------
	-- Subscribing and unsubscribing the cron service
	-- ------------------------------------------
	RegisterEventHandler(EventType.Message, "CronSubscribe", 	HandleCronSubscribe 	)
	RegisterEventHandler(EventType.Message, "CronUnSubscribe",	HandleCronUnSubscribe	)
	RegisterEventHandler(EventType.Message, "CronRemoveObject", HandleCronRemoveObject	)

	RegisterEventHandler(EventType.Message, "CronRebuildSchedule", HandleCronRebuildSchedule	)

	
	-- Self Servicing Message to rebuild the schedule at 00:01
	-- This is needed as long as we have no real calendar and only work from day to day
	RegisterEventHandler(EventType.Timer, "Midnight", HandleMidnight)


	-- Start Midnight rebuild timer
	if this:HasTimer("Midnight") then this:RemoveTimer("Midnight") end
	this:ScheduleTimerDelay(TimeSpan.FromMilliseconds(((  24-GetGameTimeOfDay())*3600*1000 ) * FACTOR), "Midnight", "Startup Sequence" ) 

	-- Debug message
	DebugMessage("Finished Cron startup sequence", true)
	end
